/*
 * layout_tiling.c - taiwins desktop tiling layout implementation
 *
 * Copyright (c) 2020 Xichen Zhou
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <string.h>
#include <stdlib.h>
#include <pixman.h>
#include <ctypes/helpers.h>
#include <ctypes/sequential.h>
#include <ctypes/tree.h>
#include <wayland-server.h>

#include "workspace.h"
#include "layout.h"
#include "xdg.h"

static const uint32_t TILINT_STATE =
	TW_XDG_VIEW_TILED_LEFT | TW_XDG_VIEW_TILED_RIGHT |
	TW_XDG_VIEW_TILED_TOP | TW_XDG_VIEW_TILED_BOTTOM;

void
emplace_tiling(const enum tw_xdg_layout_command command,
               const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
               struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops);

/******************************************************************************
 * tiling lyaout
 *****************************************************************************/

struct tiling_output;
struct tiling_view {
	//vertical split or horizental split
	bool vertical;
	float portion;
	//the interval is updated when inserting/deleting/resizing
	float interval[2];
	//you can check empty by view or check the size of the node
	struct tw_xdg_view *v;
	struct vtree_node node;
	//we have 32 layers of split
	int8_t coding[32];
	int8_t level;
	struct tiling_output *output;
};


struct tiling_output {
	//if o is empty, we don't have any outputs
	struct tiling_view *root;
	struct tw_xdg_output *output;
	struct wl_list link;
	bool used;
};

struct tiling_user_data {
	struct tiling_output outputs[32];

};

static inline struct tiling_view *
tiling_new_view(struct tw_xdg_view *v, struct tiling_output *output)
{
	struct tiling_view *tv = calloc(1, sizeof(struct tiling_view));
	vtree_node_init(&tv->node,
			offsetof(struct tiling_view, node));
	tv->v = v;
	tv->output = output;
	return tv;
}

static inline void
tiling_free_view(struct tiling_view *v)
{
	free(v);
}

static void
_free_tiling_output_view(void *data)
{
	struct tiling_view *view = data;
	free(view);
}

void
tw_xdg_layout_init_tiling(struct tw_xdg_layout *layout)
{
	tw_xdg_layout_init(layout);
	layout->user_data = calloc(1, sizeof(struct tiling_user_data));
	layout->type = LAYOUT_TILING;
	layout->command = emplace_tiling;
}

void
tw_xdg_layout_end_tiling(struct tw_xdg_layout *l)
{
	struct tiling_user_data *user_data = l->user_data;
	tw_xdg_layout_release(l);
	free(user_data);
	l->user_data = NULL;
}

/******************************************************************************
 * tiling output
 *****************************************************************************/

static inline struct tiling_output *
tiling_output_find(struct tw_xdg_layout *l, struct tw_xdg_output *o)
{
	struct tiling_user_data *user_data = l->user_data;
	return &user_data->outputs[o->idx];
}

/******************************************************************************
 * tiling view tree operations
 *****************************************************************************/

static inline struct tiling_view *
tiling_view_ith_node(struct tiling_view *parent, off_t index)
{
	return container_of(
		vtree_ith_child(&parent->node, index), struct tiling_view,
		node);
}

static int
cmp_views(const void *v, const struct vtree_node *n)
{
	const struct tiling_view *tv =
		container_of(n, const struct tiling_view, node);
	return (v == tv->v) ? 0 : -1;
}

static struct tiling_view *
tiling_view_find(struct tiling_view *root, struct tw_xdg_view *v)
{
	struct vtree_node *tnode =
		vtree_search(&root->node, v, cmp_views);
	return (tnode) ? container_of(tnode, struct tiling_view, node) : NULL;
}

//update based on portion
static inline void
tiling_update_children(struct tiling_view *parent)
{
	float leading = 0.0;
	for (int i = 0; i < parent->node.children.len; i++) {
		struct tiling_view *sv = tiling_view_ith_node(parent, i);
		sv->coding[sv->level-1] = i;
		sv->interval[0] = leading;
		sv->interval[1] = (i == (parent->node.children.len-1)) ? 1.0 :
			leading + sv->portion;
		sv->vertical = (sv->v) ? parent->vertical :
			sv->vertical;
		leading += sv->portion;
	}
}

/**
 * @brief the ONE interface to test whether it is possible to make change in
 * subtree
 */
static inline bool
is_subtree_valid(const pixman_rectangle32_t *space, const bool vertical,
		 const float portions[], const size_t len,
		 const struct tiling_output *constrain)
{
	//first lets
	float outer_size = (vertical) ? space->width : space->height;
	float inner_size = (vertical) ? space->height : space->width;
	float min_portion = (constrain->output->inner_gap * 2.0) / inner_size;

	if (constrain->output->outer_gap * 2 >= outer_size)
		return false;
	for (unsigned i = 0; i < len; i++)
		if (portions[i] < min_portion)
			return false;
	return true;
}

/* it could failed to insert the view */
static bool
tiling_view_insert(struct tiling_view *parent,
                   struct tiling_view *tv, off_t offset,
                   const pixman_rectangle32_t *parent_geo,
                   const struct tiling_output *output)
{
	double occupied = 1.0 - (double)parent->node.children.len /
		((double)parent->node.children.len+1);
	double occupied_rest = 1.0 - occupied;
	{
		//test possibility of insert
		size_t len = parent->node.children.len;
		float portions[len+1];
		for (unsigned i = 0; i < len; i++) {
			struct tiling_view *sv =
				tiling_view_ith_node(parent, i);
			portions[i] = sv->portion * occupied_rest;
		}
		portions[len] = occupied;
		if (!is_subtree_valid(parent_geo, parent->vertical,
				      portions, len+1, output))
			return false;

	}
	//re-assign all the portions
	for (int i = 0; i < parent->node.children.len; i++) {
		struct tiling_view *sv = tiling_view_ith_node(parent, i);
		sv->portion *= occupied_rest;
	}
	tv->portion = occupied;
	tv->level = parent->level+1;
	//assert just in case
	assert(tv->level <= 31);
	memcpy(tv->coding, parent->coding, sizeof(int8_t) * 32);
	tv->coding[tv->level-1] = offset;
	tv->output = parent->output;
	tv->vertical = parent->vertical;
	vtree_node_insert(&parent->node, &tv->node, offset);

	//update the subtree
	tiling_update_children(parent);
	return true;
}

static struct tiling_view *
tiling_view_erase(struct tiling_view *view)
{
	//you cannot remove a non-leaf node
	if (view->node.children.len)
		return view;
	//try to get the portion here, I don't know if removing everything it is
	//a good idea
	double rest_occupied = 1.0 - view->portion;
	off_t index = view->coding[view->level-1];
	struct tiling_view *parent = view->node.parent ?
		container_of(view->node.parent, struct tiling_view, node) :
		NULL;
	vtree_node_remove(view->node.parent, index);
	tiling_free_view(view);
	if (!parent)
		return NULL;
	//updating children info
	float leading = 0.0;
	for (int i = 0; i < parent->node.children.len; i++) {
		struct tiling_view *sv = tiling_view_ith_node(parent, i);
		sv->coding[sv->level-1] = i;
		sv->portion /= rest_occupied;
		sv->interval[0] = leading;
		sv->interval[1] = (i == (parent->node.children.len-1)) ?
			1.0 : leading + sv->portion;
		leading += sv->portion;
	}
	//if the loop above was executed, This recursive code would not run.
	//otherwise, It means this parent is a empty node:
	// 1) it is not root (it has parent).
	// 2) It does not has children.
	// 3) it has no view. Then this parent is safe to remove
	if (parent && !parent->v &&
	    !parent->node.children.len && parent->node.parent)
		return tiling_view_erase(parent);
	return parent;
}

/**
 * /brief resize a view and populate the effect on subtree
 *
 * we try to have the most complex effect resizing effect here, depends on where
 * you have your have your cursor.
 */
static bool
tiling_view_resize(struct tiling_view *view,
                   float delta_head, float delta_tail,
                   const pixman_rectangle32_t *parent_geo,
                   const struct tiling_output *output)
{
	struct tiling_view *parent = view->node.parent ?
		container_of(view->node.parent, struct tiling_view, node) :
		NULL;
	//I am the only node
	if (!parent || parent->node.children.len <= 1)
		return false;
	//deal with delta_tail, delta_head
	if (view->coding[view->level-1] == parent->node.children.len-1)
		delta_tail = 0.0;
	if (view->coding[view->level-1] == 0)
		delta_head = 0.0;
	if (delta_head == 0.0 && delta_tail == 0.0)
		return false;

	//get new portions
	float portions[parent->node.children.len];
	int index = view->coding[view->level-1];
	//space left for left views
	float occupied_rest = view->interval[0] + delta_head;
	for (int i = 0; i < index; i++) {
		struct tiling_view *tv = tiling_view_ith_node(parent, i);
		portions[i] = tv->portion * (occupied_rest /view->interval[0]);
	}
	//space for view
	portions[index] = view->portion - delta_head + delta_tail;
	//space left for right views
	float occupied_orig = 1.0 - view->interval[1];
	occupied_rest = 1 - (view->interval[1] + delta_tail);
	for (int i = index+1; i < parent->node.children.len; i++) {
		struct tiling_view *tv = tiling_view_ith_node(parent, i);
		portions[i] = tv->portion * (occupied_rest / occupied_orig);
	}
	if (!is_subtree_valid(parent_geo, parent->vertical, portions,
			      parent->node.children.len, output))
		return false;

	for (int i = 0; i < parent->node.children.len; i++) {
		struct tiling_view *tv = tiling_view_ith_node(parent, i);
		tv->portion = portions[i];
	}
	tiling_update_children(parent);
	return true;
}

/**
 * @brief shift a view in its parent list
 *
 * TODO: add this later to the layout operations
 */
void
tiling_view_shift(struct tiling_view *view, bool forward)
{
	vtree_node_shift(&view->node, forward);
	if (view->node.parent)
		tiling_update_children((struct tiling_view *)
		                       vtree_container(view->node.parent));
}

/**
 * /brief dividing a space of subtree from its parent
 */
static pixman_rectangle32_t
tiling_subtree_space(struct tiling_view *v,
                     struct tiling_view *root,
                     const pixman_rectangle32_t *space)
{
	struct tiling_view *subtree = root;
	pixman_rectangle32_t geo = *space;

	//from 0 to level - 1
	for (int i = root->level; i < v->level; i++) {
		struct vtree_node *node =
			vtree_ith_child(&subtree->node,
					v->coding[i]);
		struct tiling_view *n =
			container_of(node, struct tiling_view, node);
		if (subtree->vertical) {
			geo.y = geo.y + geo.height * n->interval[0];
			geo.height = n->portion * geo.height;
		} else {
			geo.x = geo.x + geo.width * n->interval[0];
			geo.width = n->portion * geo.width;
		}
		subtree = n;
	}
	return geo;
}

static int
tiling_arrange_subtree(struct tiling_view *subtree, pixman_rectangle32_t *geo,
                       struct tw_xdg_layout_op *data,
                       const struct tiling_output *o)
{
	//leaf
	if (subtree->v) {
		data->v = subtree->v;
		data->out.pos.x = geo->x + ((subtree->vertical) ?
					    o->output->outer_gap :
					    o->output->inner_gap);
		data->out.pos.y = geo->y + ((subtree->vertical) ?
					    o->output->inner_gap :
					    o->output->outer_gap);
		data->out.size.width =
			geo->width - 2 * ((subtree->vertical) ?
					  o->output->outer_gap :
					  o->output->inner_gap);
		data->out.size.height =
			geo->height - 2 * ((subtree->vertical) ?
					   o->output->inner_gap :
					   o->output->outer_gap);
		data->out.tile_state = TILINT_STATE;
		data->out.end = false;
		return 1;
	}
	//internal node
	int count = 0;
	for (int i = 0; i < subtree->node.children.len; i++) {
		struct vtree_node *node =
			vtree_ith_child(&subtree->node, i);
		struct tiling_view *n =
			container_of(node, struct tiling_view, node);
		pixman_rectangle32_t sub_space = *geo;
		float portion = n->interval[1] - n->interval[0];

		sub_space.x += (subtree->vertical) ?
			0 : n->interval[0] * geo->width;
		sub_space.width = (subtree->vertical) ?
			geo->width : portion * geo->width;
		sub_space.y += (subtree->vertical) ?
			n->interval[0] * geo->height : 0;
		sub_space.height = (subtree->vertical) ?
			portion * geo->height : geo->height;

		count += tiling_arrange_subtree(n, &sub_space, &data[count], o);
	}
	return count;
}

/* the launch point is based on last focused view */
static struct tiling_view*
tiling_find_launch_point(struct tw_xdg_layout *l, struct tiling_output *to,
                         struct tw_xdg_view *focused)
{
	//parent view, focused view
	struct tiling_view *pv, *fv = (focused) ?
		tiling_view_find(to->root, focused) : NULL;
	if (!fv) return to->root;
	//test if fv is root node
	pv = (fv->node.parent) ?
		container_of(fv->node.parent, struct tiling_view, node) : fv;
	//test if tv is too deep.
	pv = (pv->level >= 31) ?
		container_of(pv->node.parent, struct tiling_view, node) : pv;
	return pv;
}

/******************************************************************************
 * tiling APIs
 *****************************************************************************/
static void
tiling_add(const enum tw_xdg_layout_command command,
           const struct tw_xdg_layout_op *arg,
	   struct tw_xdg_view *v, struct tw_xdg_layout *l,
	   struct tw_xdg_layout_op *ops)
{
	//insert view based on lasted focused view
	struct tiling_output *to = tiling_output_find(l, v->output);
	//find a parent view for current layout
	struct tiling_view *pv = tiling_find_launch_point(l, to, arg->focused);
	struct tiling_output *tiling_output = pv->output;
	struct tiling_view *root = tiling_output->root;
	pixman_rectangle32_t space =
		tiling_subtree_space(pv, root,
		                     &tiling_output->output->desktop_area);

	struct tiling_view *new_view = tiling_new_view(v, tiling_output);
	//we could fail to insert
	if (tiling_view_insert(pv, new_view, 0,
			       &space, tiling_output)) {
		int count = tiling_arrange_subtree(pv, &space,
		                                   ops, tiling_output);
		ops[count].out.end = true;
	} else {
		tiling_free_view(new_view);
		ops[0].out.end = true;
	}
}

static void
tiling_del(const enum tw_xdg_layout_command command,
           const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
           struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	struct tiling_output *output = tiling_output_find(l, v->output);
	struct tiling_view *view = output ?
		tiling_view_find(output->root, v) : NULL;
	struct tiling_view *parent = view ? tiling_view_erase(view) : NULL;
	if (parent) {
		pixman_rectangle32_t space =
			tiling_subtree_space(parent, output->root,
					     &output->output->desktop_area);
		int count = tiling_arrange_subtree(parent, &space,
		                                   ops, output);
		ops[count].out.end = true;
	} else
		ops[0].out.end = true;

}

static inline enum wl_shell_surface_resize
_tiling_resize_correct_edge(struct tiling_view *view)
{
	bool last = view->interval[1] >= 0.999;
	return (last) ? WL_SHELL_SURFACE_RESIZE_TOP_LEFT :
		WL_SHELL_SURFACE_RESIZE_BOTTOM_RIGHT;
}

static void
_tiling_resize(const struct tw_xdg_layout_op *arg, struct tiling_view *view,
	       struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops,
               bool force_update)
{
	enum wl_shell_surface_resize edge = _tiling_resize_correct_edge(view);
	struct tiling_output *tiling_output = view->output;
	struct tiling_view *parent = container_of(view->node.parent,
						  struct tiling_view, node);
	pixman_rectangle32_t space =
		tiling_subtree_space(parent, tiling_output->root,
				     &tiling_output->output->desktop_area);
	//head, tail
	float ph = 0.0, pt = 0.0;
	if (parent->vertical) {
		ph = (edge & WL_SHELL_SURFACE_RESIZE_TOP) ?
			arg->in.dy / space.height : 0.0;
		pt = (edge & WL_SHELL_SURFACE_RESIZE_BOTTOM) ?
			arg->in.dy / space.height : 0.0;
	} else {
		ph = (edge & WL_SHELL_SURFACE_RESIZE_LEFT) ?
			arg->in.dx / space.width : 0.0;
		pt = (edge & WL_SHELL_SURFACE_RESIZE_RIGHT) ?
			arg->in.dx / space.width : 0.0;
	}
	bool resized = tiling_view_resize(view, ph, pt, &space,
					  tiling_output);
	//try to resize the parent->parent,
	struct tiling_view *gparent = (parent->node.parent) ?
		container_of(parent->node.parent, struct tiling_view, node) :
		NULL;
	ops[0].out.end = true;
	if (gparent) {
		_tiling_resize(arg, parent, l, ops, resized || force_update);
	} else if (resized || force_update) {
		int count = tiling_arrange_subtree(parent, &space,
		                                   ops, tiling_output);
		ops[count].out.end = true;
	}
}

static void
tiling_resize(const enum tw_xdg_layout_command command,
              const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
              struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	struct tiling_output *tiling_output = tiling_output_find(l, v->output);
	struct tiling_view *view = tiling_view_find(tiling_output->root, v);
	_tiling_resize(arg, view, l, ops, false);
}

/**
 * @brief do a vertical/horizental split
 */
static void
_tiling_split(const enum tw_xdg_layout_command command,
              const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
              struct tw_xdg_layout *l, bool vertical,
              struct tw_xdg_layout_op *ops)
{
	struct tiling_output *tiling_output = tiling_output_find(l, v->output);
	struct tiling_view *view = tiling_view_find(tiling_output->root, v);
	struct tiling_view *parent = container_of(view->node.parent,
						  struct tiling_view, node);

	//test if the view is the only child. So we do not need to split
	if (parent->node.children.len <= 1) {
		parent->vertical = vertical;
		ops[0].out.end = true;
		return;
	}

	pixman_rectangle32_t space =
		tiling_subtree_space(view, tiling_output->root,
				     &tiling_output->output->desktop_area);
	struct tiling_view *new_view = tiling_new_view(v, tiling_output);
	view->v = NULL;
	view->vertical = vertical;
	tiling_view_insert(view, new_view, 0, &space, tiling_output);
	int count = tiling_arrange_subtree(view, &space, ops, tiling_output);
	ops[count].out.end = true;
}

static void
tiling_vsplit(const enum tw_xdg_layout_command command, const struct tw_xdg_layout_op *arg,
	      struct tw_xdg_view *v, struct tw_xdg_layout *l,
	      struct tw_xdg_layout_op *ops)
{
	_tiling_split(command, arg, v, l, true, ops);
}

static void
tiling_hsplit(const enum tw_xdg_layout_command command,
              const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
              struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	_tiling_split(command, arg, v, l, false, ops);
}

static void
tiling_merge(const enum tw_xdg_layout_command command,
             const struct tw_xdg_layout_op *arg,
             struct tw_xdg_view *v, struct tw_xdg_layout *l,
             struct tw_xdg_layout_op *ops)
{
	//remove current view and then insert at grandparent list
	struct tiling_output *tiling_output = tiling_output_find(l, v->output);
	struct tiling_view *view = tiling_view_find(tiling_output->root, v);
	struct tiling_view *parent = container_of(view->node.parent,
						  struct tiling_view, node);
	struct tiling_view *gparent = parent->node.parent ?
		container_of(parent->node.parent, struct tiling_view, node) :
		NULL;
	//if we are not
	if (view->node.children.len || !gparent) {
		ops[0].out.end = true;
		return;
	}
	pixman_rectangle32_t space =
		tiling_subtree_space(gparent, tiling_output->root,
				     &tiling_output->output->desktop_area);

	tiling_view_erase(view);
	view = tiling_new_view(v, tiling_output);
	//TODO deal with the case that it cannot insert
	tiling_view_insert(gparent, view, 0, &space, tiling_output);
	int count = tiling_arrange_subtree(gparent, &space, ops, tiling_output);
	ops[count].out.end = true;
}

/**
 * @brief toggle vertical
 */
static void
tiling_toggle(const enum tw_xdg_layout_command command,
              const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
              struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	struct tiling_output *tiling_output = tiling_output_find(l, v->output);
	struct tiling_view *view = tiling_view_find(tiling_output->root, v);
	struct tiling_view *parent = container_of(view->node.parent,
						  struct tiling_view, node);
	pixman_rectangle32_t space =
		tiling_subtree_space(parent, tiling_output->root,
				     &tiling_output->output->desktop_area);
	if (parent) {
		parent->vertical = !parent->vertical;
		int count = tiling_arrange_subtree(parent, &space,
		                                   ops, tiling_output);
		ops[count].out.end = true;
	}
}

static void
tiling_add_output(const enum tw_xdg_layout_command command,
                  const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
                  struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	struct tiling_user_data *user_data = l->user_data;
	struct tw_xdg_output *xdg_output = arg->in.o;
	struct tiling_output *output = &user_data->outputs[xdg_output->idx];

	output->output = xdg_output;
	//setup the first node
	output->root = tiling_new_view(NULL, output);
	output->root->level = 0;
	output->root->portion = 1.0;
	output->root->vertical = false;
	ops[0].out.end = true;
}

static void
tiling_rm_output(const enum tw_xdg_layout_command command,
                 const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
                 struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	struct tiling_user_data *user_data = l->user_data;
	struct tiling_output *output = &user_data->outputs[arg->in.o->idx];
	vtree_destroy(&output->root->node, _free_tiling_output_view);
	ops[0].out.end = true;
 }

static void
tiling_resize_output(const enum tw_xdg_layout_command command,
                     const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
                     struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	struct tiling_user_data *user_data = l->user_data;
	struct tiling_output *output = &user_data->outputs[arg->in.o->idx];

	int count = tiling_arrange_subtree(output->root,
	                                   &output->output->desktop_area,
	                                   ops, output);
	ops[count].out.end = true;
}

void
emplace_tiling(const enum tw_xdg_layout_command command,
               const struct tw_xdg_layout_op *arg, struct tw_xdg_view *v,
               struct tw_xdg_layout *l, struct tw_xdg_layout_op *ops)
{
	//make a dummy one now
	struct placement_node {
		enum tw_xdg_layout_command command;
		tw_xdg_layout_fun_t fun;
	};

	static struct placement_node t_ops[] = {
		{DPSR_focus, tw_xdg_layout_emplace_noop},
		{DPSR_add, tiling_add},
		{DPSR_del, tiling_del},
		{DPSR_deplace, tw_xdg_layout_emplace_noop},
		{DPSR_toggle, tiling_toggle},
		{DPSR_resize, tiling_resize},
		{DPSR_vsplit, tiling_vsplit},
		{DPSR_hsplit, tiling_hsplit},
		{DPSR_merge, tiling_merge},
		{DPSR_output_add, tiling_add_output},
		{DPSR_output_rm, tiling_rm_output},
		{DPSR_output_resize, tiling_resize_output},
	};
	assert(t_ops[command].command == command);
	t_ops[command].fun(command, arg, v, l, ops);
}
